#include <windows.h>
#include <tchar.h>
#include "calc.h"
#include "regex.h"

#include <vector>
#include <stack>

#define SPLIT_TOKEN_ERROR ((int) 1)
#define UNKNOWN_TOKEN_ERROR ((int) 2)
#define UNKNOWN_OPERATION_ERROR ((int) 3)
#define STACK_SIZE_ERROR ((int) 4)

typedef enum _RPN_TYPE {
	DIGIT,
	ADD,
	SUB,
	MUL,
	DIV,
	BRACKET,
} RPN_TYPE;

typedef struct _RPN_DATA {
	RPN_TYPE type;
	int nest;
	double digit;
} RPN_DATA;

void CreateRpnData(LPTSTR exp, int *nest, RPN_DATA *lpData) {
	RPN_TYPE type;
	double digit = 0;
	switch (exp[0]) {
	case _T('+'): type = ADD; break;
	case _T('-'): type = SUB; break;
	case _T('*'): type = MUL; break;
	case _T('/'): type = DIV; break;
	case _T('('): type = BRACKET; (*nest)++; break;
	case _T(')'): type = BRACKET; (*nest)--; break;
	default:
		type = DIGIT;
		digit = _tcstod(exp, NULL);
		break;
	}
	if (type < 0) return;
	lpData->type = type;
	lpData->nest = *nest;
	lpData->digit = digit;
}

int GetPriority(RPN_DATA &data) {
	switch (data.type) {
	case ADD:
	case SUB:
		return 2;
	case MUL:
	case DIV:
		return 3;
	case DIGIT:
		return 9;
	default:
		return -1;
	}
}

int ComparePriority(RPN_DATA &data1, RPN_DATA &data2) {
	int result = 0;
	result = data1.nest - data2.nest;
	if (result == 0) {
		result = GetPriority(data1) - GetPriority(data2);
	}
	return (result > 0 ? 1 : (result < 0 ? -1 : 0));
}

// ȉ̕ϊs
// QoCgΉPoCgɕϊ
// 󔒂
// gpdelete[]邱
LPTSTR ConvertString(LPCTSTR lpCmdLine) {
	size_t len = _tcslen(lpCmdLine);
	LPTSTR result = new TCHAR[len + 1];
	LPTSTR buf = result;
	ZeroMemory(result, sizeof(TCHAR) * (len + 1));
	for (size_t i = 0; i < len; i++) {
		switch (lpCmdLine[i]) {
		case _T(' '):
		case _T('\t'):
		case _T('@'):
			// 󔒗ނ
			break;
		case _T('0'): case _T('1'): case _T('2'): case _T('3'): case _T('4'):
		case _T('5'): case _T('6'): case _T('7'): case _T('8'): case _T('9'):
		case _T('.'):
		case _T('+'): case _T('-'): case _T('*'): case _T('/'):
		case _T('('): case _T(')'):
			// 1oCg͂̂܂܃Zbg
			*(buf++) = lpCmdLine[i];
			break;
		case _T('O'): case _T('P'): case _T('Q'): case _T('R'): case _T('S'):
		case _T('T'): case _T('U'): case _T('V'): case _T('W'): case _T('X'):
			//QoCglPoCglɕϊ
			*(buf++) = (lpCmdLine[i] - _T('O') + _T('0'));
			break;
		case _T('D'):
			*(buf++) = _T('.');
			break;
		case _T('{'):
			*(buf++) = _T('+');
			break;
		case _T('['):
		case _T('|'):
		case _T('\'):
		case _T(']'):
			*(buf++) = _T('-');
			break;
		case _T(''):
		case _T('~'):
			*(buf++) = _T('*');
			break;
		case _T('^'):
		case _T(''):
			*(buf++) = _T('/');
			break;
		case _T('i'):
			*(buf++) = _T('(');
			break;
		case _T('j'):
			*(buf++) = _T(')');
			break;
		default:
			throw UNKNOWN_TOKEN_ERROR;
		}
	}
	return result;
}

// lZt|[hL@Ƀp[X
void ParseRPN(LPCTSTR lpCmdLine, std::vector<RPN_DATA> &result) {
	CRegex re;
	std::stack<RPN_DATA> stack;
	re.SetPattern(_T("^(\\+|-|\\*|/|\\(|\\)|[0-9]+(\\.[0-9]*)?)"));
	LPTSTR str = ConvertString(lpCmdLine);
	LPTSTR next = str;
	int nest = 0;
	while (next != NULL) {
		if (re.Search(next) == 0) break;
		LPTSTR exp = re.GetMatch(1);
		RPN_DATA data;
		CreateRpnData(exp, &nest, &data);
		switch (data.type) {
		case DIGIT:
			result.push_back(data);
			break;
		case BRACKET:
			//ʂ͒ǉȂ
			break;
		default:
			while (stack.size() > 0 && ComparePriority(stack.top(), data) > 0) {
				result.push_back(stack.top());
				stack.pop();
			}
			stack.push(data);
			break;
		}
		next = re.GetNext();
	}
	while (stack.size() > 0) {
		result.push_back(stack.top());
		stack.pop();
	}
	delete[] str;
}

// t|[h̃f[^vZ
double CalcRPN(std::vector<RPN_DATA> &rpn) {
	std::stack<double> stack;
	for (std::vector<RPN_DATA>::iterator it = rpn.begin(); it != rpn.end(); it++) {
		if (it->type == DIGIT) {
			stack.push(it->digit);
		} else {
			if (stack.size() < 2) throw STACK_SIZE_ERROR;
			double op2 = stack.top();
			stack.pop();
			double op1 = stack.top();
			stack.pop();
			switch (it->type) {
			case ADD:
				op1 += op2;
				break;
			case SUB:
				op1 -= op2;
				break;
			case MUL:
				op1 *= op2;
				break;
			case DIV:
				op1 /= op2;
				break;
			default:
				throw UNKNOWN_OPERATION_ERROR;
			}
			stack.push(op1);
		}
	}
	if (stack.size() != 1) throw STACK_SIZE_ERROR;
	return stack.top();
}

BOOL Calc(LPCTSTR lpCmdLine, LPTSTR lpResult, SIZE_T size) {
	std::vector<RPN_DATA> rpn;
	try {
		ParseRPN(lpCmdLine, rpn);
		double result = CalcRPN(rpn);
		if ((int) result == result) {
			_sntprintf(lpResult, size, _T("%d"), (int) result);
		} else {
			_sntprintf(lpResult, size, _T("%f"), result);
		}
	} catch (int e) {
		_sntprintf(lpResult, size, _T("vZł܂B[%s](error code: %d)"), lpCmdLine, e);
	}
	return TRUE;
}
